/*////////////////////////////////////////////////////////////////////////
Copyright (c) 1996-1999 Yutaka Sato
Copyright (c) 1996-1999 Electrotechnical Laboratry (ETL), AIST, MITI

Permission to use, copy, and distribute this material for any purpose
and without fee is hereby granted, provided that the above copyright
notice and this permission notice appear in all copies.
ETL MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY OF THIS
MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS", WITHOUT ANY EXPRESS
OR IMPLIED WARRANTIES.
/////////////////////////////////////////////////////////////////////////
Content-Type:	program/C; charset=US-ASCII
Program:	socks.c
Author:		Yutaka Sato <ysato@etl.go.jp>
Description:

	SOCKS=host[:[port][/version][:dstHostList[:srcHostList]]]

History:
	991119	extracted from socks4.c
//////////////////////////////////////////////////////////////////////#*/
#include <stdio.h>
#include "delegate.h"
extern char *gethostaddr();

#define SOCKS_PORT	1080
#define SOCKS_CONNECT	1
#define SOCKS_BIND	2
#define SOCKS_ACCEPT	8	/* pseudo command */

typedef struct {
	char	*s_host;
	int	 s_port;
	int	 s_ver;
	int	 s_rslv; /* V4:delegate resolution, V5:resolv in mylsef */
	int	 s_sock;
	int	 s_acl;
} SocksServer;

typedef struct {
  SocksServer  *v_serv;
	int	v_addr;
	int	v_port;
	VSAddr	v_local; /* source address for the destination */
} ViaSocks;
#define NVIAS 64

typedef struct {
	ViaSocks se_viaSocks[NVIAS];
	char	se_username[64];
	SocksServer se_sockservs[8];
	int	se_bind_ver;
	int	se_resolv_bymyself;
} SocksEnv;

static SocksEnv *socksEnv;
#define viaSocks	socksEnv->se_viaSocks
#define username	socksEnv->se_username
#define sockservs	socksEnv->se_sockservs
#define bind_ver	socksEnv->se_bind_ver
#define resolv_bymyself(host)	socksEnv->se_resolv_bymyself

minit_socks()
{
	if( socksEnv == 0 )
		socksEnv = NewStruct(SocksEnv);
}

/*
 *	SOCKS CLIENT
 */

extern char *hostmatch_asisaddr;
static socks_init(_,svhost,svportver,dsts,srcs)
	void *_;
	char *svhost,*svportver,*dsts,*srcs;
{	int sx;
	char *opts,*host1;
	char key[32],map[1024];
	int port;
	int ver = 5;

	if( svhost == 0 || svhost[0] == 0 )
		return;

	port = 0;
	opts = NULL;

	if( svportver ){
		port = atoi(svportver);
		if( opts = strchr(svportver,'/') )
			opts++;
	}
	if( port == 0 )
		port = SOCKS_PORT;
	if( opts != NULL ){
		if( strcmp(opts,"-4") == 0 )
			ver = 4;
		if( strcmp(opts,"-r") == 0 )
			socksEnv->se_resolv_bymyself = 1;
	}

	for( sx = 0; host1 = sockservs[sx].s_host; sx++ ){
		if( host1[0] == 0 )
			break;
		if( sockservs[sx].s_port == port )
		if( hostcmp(sockservs[sx].s_host,svhost) == 0 )
			return;
	}

	sockservs[sx].s_host = stralloc(svhost);
	sockservs[sx].s_port = port;
	sockservs[sx].s_ver = ver;

	sprintf(key,"SOCKS-%d",sx);
/*
	sprintf(map,"{%s:%d/-%d}:*:%s:%s",
		svhost,port,ver,dsts?dsts:".socksdst",srcs?srcs:"*");
*/
	sprintf(map,"{%s:%d/-%d}:*:%s,%s:%s,%s",svhost,port,ver,
			hostmatch_asisaddr,dsts?dsts:".socksdst",
			hostmatch_asisaddr,srcs?srcs:"*");
	sockservs[sx].s_acl = scan_CMAP2(_,key,map);
}
static matchConn(Conn,sx)
	Connection *Conn;
{	char key[32],map[128];
	int found;

	sprintf(key,"SOCKS-%d",sx);
	CTX_pushClientInfo(Conn);
	found = find_CMAP(Conn,key,map);
	HL_popClientInfo();
	return 0 <= found;
}
static matchConnX(Conn,sx,dstproto,dsthost,dstport)
	Connection *Conn;
	char *dstproto,*dsthost;
{	int sav,found;

	sav = 0;
	if( streq(DFLT_PROTO,"socks") && REAL_PROTO[0] == 0 ){
		strcpy(REAL_PROTO,dstproto);
		wordScan(dsthost,REAL_HOST);
		REAL_PORT = dstport;
		sav = 1;
	}
	found = matchConn(Conn,sx);
	if( sav ){
		REAL_PROTO[0] = 0;
		REAL_HOST[0] = 0;
		REAL_PORT = 0;
	}
	return found;
}
static tobeViaSocks(Conn)
	Connection *Conn;
{	int sx;

	for( sx = 0; sockservs[sx].s_host; sx++ ){
		if( matchConn(Conn,sx) )
			return 1;
	}
	return 0;
}
GetViaSocks(Conn,host,port)
	Connection *Conn;
	char *host;
{
	return VSA_getViaSocks(Conn,host,port,NULL);
}
VSA_getViaSocks(Conn,host,port,vlocal)
	Connection *Conn;
	char *host;
	VSAddr *vlocal;
{	Connection toConn;
	int match,sx;

	/*
	... it should be controlled by tobeViaSocks() ...
	if( hostcmp(host,"localhost") == 0 ){
		Verbose("##NOT ViaSocks-A(localhost)## %s:%d\n",host,port);
		return 0;
	}
	*/
	if( getViaSocks(host,port,vlocal) ){
		sv1log("##ViaSocks-A(by cache)## %s:%d\n",host,port);
		return 1;
	}
	toConn = *Conn;
	set_realserver(&toConn,DST_PROTO,host,port);
	if( tobeViaSocks(&toConn) ){
		sv1log("##ViaSocks-B(by rule)## %s:%d\n",host,port);
		if( vlocal ) bzero(vlocal,sizeof(VSAddr));
		return 1;
	}
	Verbose("##NOT ViaSocks-B## %s:%d\n",host,port);
	return 0;
}
scan_SOCKS(_,socks)
	void *_;
	char *socks;
{	char socksb[1024],*sv[4];

	strcpy(socksb,socks);
	sv[0] = sv[1] = sv[2] = sv[3] = 0;
	stoV(socksb,3,sv,':');
	socks_init(_,sv[0],sv[1],sv[2],sv[3]);
}
socks_addservers(){
	int sx;
	char *host;
	char *addr;

	for( sx = 0; host = sockservs[sx].s_host; sx++ )
	if( *host )
	{
		addr = gethostaddr(host);
		if( addr == NULL )
			addr = host;
		SOCKS_addserv("*",0,addr,sockservs[sx].s_port);
		/*
		SOCKS_addserv("*",0,gethostaddr(host),sockservs[sx].s_port);
		*/
	}
	return sx;
}

static socks_start(Conn,sv,command,host,port,user,rhost,rport,xaddr)
	Connection *Conn;
	SocksServer *sv;
	char *host,*user,*rhost;
	int *rport;
	char *xaddr;
{	unsigned char obuf[128],i,*ap;
	int pc,wcc,rep;
	char *auth,userb[128],*dp,*pass;
	char authb[256];
	char *skhost = sv->s_host;
	int skport = sv->s_port;
	int ver = sv->s_ver;
	int sock = sv->s_sock;

	if( ver == 5 ){

if( strchr(host,'%') ){
	char xhost[512],*addr;
	sv1log("## unescape host-name for SocksV5: %s\n",host);
	if( addr = gethostaddr(host) ){
		strcpy(xhost,addr);
		if( xaddr ) strcpy(xaddr,xhost);
	}else	nonxalpha_unescape(host,xhost,0);
	host = xhost;
}
		/* pass resolved address to SocksV5 client to know which of
		 * multiple IP addresses of the target host is used fot the
		 * connection and to pass it to setViaSocks().
		 */
		/* but this will cause problem when (especially non-FTP) user
		 * expects name resolusions to be done in SocksV5 server ...
		 */
		if( !isinetAddr(host) )
		if( command == SOCKS_CONNECT && xaddr != NULL )
		if( port == 21	/* should be switched using potocol name or
				 * explicit configuration ...
				 */
		 || sv->s_rslv
		 || resolv_bymyself(host) )
		{	char *addr;
			char addrb[128];
			if( addr = gethostaddr(host) ){
				sv1log("resolved %s[%s]\n",host,addr);
				strcpy(xaddr,addr);
/*
				host = addr;
*/
				/* addr can be overwritten in get_MYAUTH() */
				strcpy(addrb,addr);
				host = addrb;
			}
		}
		pass = NULL;
		if( get_MYAUTH(Conn,authb,"socks",skhost,skport) )
			auth = authb;
		else	auth = getenv("SOCKS_AUTH");
		if( auth ){
			dp = wordscanY(auth,userb,sizeof(userb),"^:");
			user = userb;
			if( *dp == ':' )
				pass = dp + 1;
			else	pass = dp;
		}
		return SOCKS_startV5(sock,command,host,port,user,pass,rhost,rport);
	}

	if( resolv_bymyself(host) == 0 && sv->s_rslv == 0 ){
		char *addr;
		addr = gethostaddr(host);
		if( addr == NULL ){
			sv1log("Don't try Socks for unknown host\n");
			return -1;
		}
		host = addr;
	}
	return
	SocksV4_clientStart(sock,command,host,port,user,rhost,rport,xaddr);
}

static setViaSocksX(sv,addr,port,vlocal)
	SocksServer *sv;
	unsigned int addr;
	VSAddr *vlocal;
{	ViaSocks *vs;
	int vsx;

	if( addr == -1 )
		return;

	for( vsx = 0; vsx < NVIAS; vsx++ ){
		vs = &viaSocks[vsx];
		if( vs->v_addr == 0 ){
			vs->v_serv = sv;
			vs->v_addr = addr;
			vs->v_port = port;
			vs->v_local = *vlocal;
			break;
		}
		if( vs->v_addr == addr )
			break;
	}
}
static setViaSocks(sv,host,port,vlocal)
	SocksServer *sv;
	char *host;
	VSAddr *vlocal;
{
	setViaSocksX(sv,gethostint_nboV4(host),port,vlocal);
	setViaSocksX(sv,gethostintMin(host),port,vlocal);
	setViaSocksX(sv,_inet_addrV4(host),port,vlocal);
}
static SocksServer *getViaSocksX(addr,port,vlocal)
	unsigned int addr;
	VSAddr *vlocal;
{	ViaSocks *vs;
	int vsx;

	if( addr == -1 )
		return 0;

	for( vsx = 0; vsx < NVIAS; vsx++ ){
		vs = &viaSocks[vsx];
		if( vs->v_addr == 0 )
			break;
		if( vs->v_addr == addr ){
			if( vlocal ) *vlocal = vs->v_local;
			return vs->v_serv;
		}
	}
	return 0;
}
static getViaSocks(host,port,vlocal)
	char *host;
	VSAddr *vlocal;
{
	if( getViaSocksX(gethostint_nboV4(host),port,vlocal) ) return 1;
	if( getViaSocksX(gethostintMin(host),port,vlocal) ) return 1;
	if( getViaSocksX(_inet_addrV4(host),port,vlocal) ) return 1;
	return 0;
}

static socks_connect(Conn,sv,host,port,user,rhost,rport)
	Connection *Conn;
	SocksServer *sv;
	char *host,*user,*rhost;
	int *rport;
{	char addr[64];
	VSAddr vlocal;

	if( host == NULL || port == 0 )
		return -1;

	*addr = 0;
	if( socks_start(Conn,sv,SOCKS_CONNECT,host,port,user,rhost,rport,addr))
		return -1;

	if( rhost && rport )
		VSA_atosa(&vlocal,*rport,rhost);
	else	bzero(&vlocal,sizeof(VSAddr));
	setViaSocks(sv,host,port,&vlocal);
	if( *addr != 0 )
		setViaSocks(sv,addr,port,&vlocal);
	return 0;
}

static socks_bind(Conn,sv,host,port,user,rhost,rport)
	Connection *Conn;
	SocksServer *sv;
	char *host,*user,*rhost;
	int *rport;
{
	bind_ver = sv->s_ver;
	return socks_start(Conn,sv,SOCKS_BIND,host,port,user,rhost,rport,NULL);
}
static bindsocks(Conn,sv,dsthost,dstport,uname,rhost,rport)
	Connection *Conn;
	SocksServer *sv;
	char *dsthost,*uname;
	char *rhost;
	int *rport;
{	int sock;

	sock = OpenServer("BindViaSocks","socks",sv->s_host,sv->s_port);
	if( sock < 0 )
		return -1;

	sv->s_sock = sock;
	if( socks_bind(Conn,sv,dsthost,dstport,uname,rhost,rport) == 0 )
		return sock;
	close(sock);
	return -1;
}
bindViaSocks(Conn,dsthost,dstport,rhost,rport)
	void *Conn;
	char *dsthost,*rhost;
	int *rport;
{	int sx,sock;
	SocksServer *sv;

	getUsernameCached(getuid(),username);
	sv1log("bindViaSocks(%s:%d)[%s]\n",dsthost,dstport,username);

	if( sv = getViaSocksX(gethostint_nboV4(dsthost),dstport,NULL) ){
		sock = bindsocks(Conn,sv,dsthost,dstport,username,rhost,rport);
		if( 0 <= sock )
			return sock;
	}

	for( sx = 0; sockservs[sx].s_host; sx++ ){
		/*
		if( !matchConn(Conn,sx) )
		*/
		if( !matchConnX(Conn,sx,"tcprelay",dsthost,dstport) )
			continue;

		sv = &sockservs[sx];
		if( sv->s_host[0] == 0 || sv->s_port == 0 )
			break;
		sock = bindsocks(Conn,sv,dsthost,dstport,username,rhost,rport);
		if( 0 < sock )
			return sock;
	}
	return -1;
}
acceptViaSocks(sock,rhost,rport)
	char *rhost;
	int *rport;
{	int rep;

	if( bind_ver == 5 )
		return SOCKS_recvResponseV5(sock,SOCKS_ACCEPT,rhost,rport);
	else	return SocksV4_acceptViaSocks(sock,rhost,rport);
}
connectViaSocksX(Conn,skhost,skport,opts,dsthost,dstport)
	Connection *Conn;
	char *skhost,*opts,*dsthost;
{	static SocksServer svb;
	SocksServer *sv = &svb;
	char rhost[128];
	int rport;

	sv->s_sock = OpenServer("ConnectViaSocks","socks",skhost,skport);
	sv->s_host = stralloc(skhost);
	sv->s_port = skport;
	sv->s_rslv = 0;
	if( strstr(opts,"-4") )
		sv->s_ver = 4;
	else	sv->s_ver = 5;
	if( strstr(opts,"-r") )
		sv->s_rslv = 1;

	if( 0 <= sv->s_sock ){
		getUsernameCached(getuid(),username);
	if( socks_connect(Conn,sv,dsthost,dstport,username,rhost,&rport)==0 )
			return sv->s_sock;
		close(sv->s_sock);
	}
	return -1;
}
static connsocks(Conn,sv,dsthost,dstport,uname,rhost,rport)
	Connection *Conn;
	SocksServer *sv;
	char *dsthost,*uname;
	char *rhost;
	int *rport;
{	int sock;

	sock = OpenServer("ConnectViaSocks","socks",sv->s_host,sv->s_port);
	if( sock < 0 )
		return -1;

	sv->s_sock = sock;
	if( socks_connect(Conn,sv,dsthost,dstport,uname,rhost,rport) == 0 )
		return sock;
	close(sock);
	return -1;
}
connectViaSocks(Conn,dsthost,dstport,rhost,rport)
	Connection *Conn;
	char *dsthost;
	char *rhost;
	int *rport;
{	int sx,sock;
	SocksServer *sv;

	getUsernameCached(getuid(),username);
	if( sv = getViaSocksX(gethostint_nboV4(dsthost),dstport,NULL) ){
		sock = connsocks(Conn,sv,dsthost,dstport,username,rhost,rport);
		if( 0 <= sock )
			return sock;
	}

	for( sx = 0; sockservs[sx].s_host; sx++ ){
		if( !matchConn(Conn,sx) )
			continue;

		sv = &sockservs[sx];
		if( sv->s_host[0] == 0 || sv->s_port == 0 )
			break;

		sock = connsocks(Conn,sv,dsthost,dstport,username,rhost,rport);
		if( 0 <= sock )
			return sock;
	}
	return -1;
}
ConnectViaSocks(Conn,relay_input)
	Connection *Conn;
{	int sock;
	char rhost[128];
	int rport;

	if( Conn->co_mask & CONN_NOSOCKS )
		return -1;
	if( 0 <= (sock = connectViaSocks(Conn,DST_HOST,DST_PORT,rhost,&rport)) )
		initConnected(Conn,sock,relay_input);
	return sock;
}

/*
 *	SOCKS SERVER
 */
service_socks(Conn)
	Connection *Conn;
{	unsigned char ibuf[16];

	if( recvPeekTIMEOUT(FromC,ibuf,1) != 1 )
		return;

	if( !source_permittedX(Conn) ){
		char shost[256];
		getClientHostPort(Conn,shost);
		daemonlog("F","E-P: No permission: \"%s\" is not allowed (%s)\n",
			shost,Conn->reject_reason);
		/* send appropriate error message depending on client ver. */
		return;
	}

	if( ibuf[0] == 5 ){
		extern int IO_TIMEOUT;
		socks_addservers();
		return SOCKS_serverV5(Conn,FromC,ToC,IO_TIMEOUT*1000);
	}
	SocksV4_server(Conn,ToS,FromS,FromC,ToC);
}
isSocksConnect(pack,leng,ver,addr,port,user)
	unsigned char *pack;
	int *ver;
	char *addr;
	int *port;
	unsigned char **user;
{
	return SocksV4_isConnectReuqest(pack,leng,ver,addr,port,user);
}
